/*#define INT9_TEST 1*/

#ifdef INT9_TEST
#include <stdio.h>
#include <stdlib.h>
#endif

#ifndef _WIN32
#include <dos.h>
#else
#include "win32mod.h"
#endif

#include "keyboard.h"

#define MAX_KEY 0x53
static char *lKeyNames[MAX_KEY + 1] = {
/* 0x00 */ "<None>",
/* 0x01 */ "Escape",
/* 0x02 */ "1",
/* 0x03 */ "2",
/* 0x04 */ "3",
/* 0x05 */ "4",
/* 0x06 */ "5",
/* 0x07 */ "6",
/* 0x08 */ "7",
/* 0x09 */ "8",
/* 0x0A */ "9",
/* 0x0B */ "0",
/* 0x0C */ "-",
/* 0x0D */ "+",
/* 0x0E */ "Backspace",
/* 0x0F */ "Tab",
/* 0x10 */ "Q",
/* 0x11 */ "W",
/* 0x12 */ "E",
/* 0x13 */ "R",
/* 0x14 */ "T",
/* 0x15 */ "Y",
/* 0x16 */ "U",
/* 0x17 */ "I",
/* 0x18 */ "O",
/* 0x19 */ "P",
/* 0x1A */ "{",
/* 0x1B */ "}",
/* 0x1C */ "Enter",
/* 0x1D */ "Control",
/* 0x1E */ "A",
/* 0x1F */ "S",
/* 0x20 */ "D",
/* 0x21 */ "F",
/* 0x22 */ "G",
/* 0x23 */ "H",
/* 0x24 */ "J",
/* 0x25 */ "K",
/* 0x26 */ "L",
/* 0x27 */ ":",
/* 0x28 */ "\"",
/* 0x29 */ "~",
/* 0x2A */ "Left Shift",
/* 0x2B */ "\\",
/* 0x2C */ "Z",
/* 0x2D */ "X",
/* 0x2E */ "C",
/* 0x2F */ "V",
/* 0x30 */ "B",
/* 0x31 */ "N",
/* 0x32 */ "M",
/* 0x33 */ ",",
/* 0x34 */ ".",
/* 0x35 */ "/",
/* 0x36 */ "Right Shift",
/* 0x37 */ "*",
/* 0x38 */ "Alt",
/* 0x39 */ "Space",
/* 0x3A */ "Caps Lock",
/* 0x3B */ "F1",
/* 0x3C */ "F2",
/* 0x3D */ "F3",
/* 0x3E */ "F4",
/* 0x3F */ "F5",
/* 0x40 */ "F6",
/* 0x41 */ "F7",
/* 0x42 */ "F8",
/* 0x43 */ "F9",
/* 0x44 */ "F10",
/* 0x45 */ "Num Lock",
/* 0x46 */ "Scroll Lock",
/* 0x47 */ "Home",
/* 0x48 */ "Up Arrow",
/* 0x49 */ "Page Up",
/* 0x4A */ "Grey -",
/* 0x4B */ "Left Arrow",
/* 0x4C */ "Center",
/* 0x4D */ "Right Arrow",
/* 0x4E */ "Grey +",
/* 0x4F */ "End",
/* 0x50 */ "Down Arrow",
/* 0x51 */ "Page Down",
/* 0x52 */ "Insert",
/* 0x53 */ "Delete"
};

#ifndef _WIN32

static unsigned char lPauseSeq[] = {0xE1, 0x1D, 0x45, 0xE1, 0x9D, 0xC5, 0x00};
static unsigned char nPauseNum;
static unsigned short nFlag;
static unsigned char  bExtKey;
static unsigned short nLastKey;
#ifdef __cplusplus
#define __CPPARGS ...
#else
#define __CPPARGS
#endif
typedef void interrupt (*VECTPROC)(__CPPARGS);

#define KEY_RELEASED 0x8000
void interrupt Int9Hook(__CPPARGS) {
unsigned char b, v;
  b = inp(0x60);
  v = inp(0x61);
  outp(0x61, v | 0x80);
  outp(0x61, v);
  outp(0x20, 0x20);
  if (b == lPauseSeq[nPauseNum]) {
    nPauseNum++;
    if (!lPauseSeq[nPauseNum]) {
      nPauseNum = 0;
      /* ignore Pause key */
    }
  } else {
    nPauseNum = 0;
    /* extended key */
    if (b >= 0xE0) {
      /* how much keys in the sequence */
      bExtKey = b - 0xE0 + 1;
      nFlag = bExtKey;
      nFlag <<= 8;
    } else {
      nFlag &= bExtKey ? 0xFF00 : 0x0000;
      /* skip this key */
      bExtKey -= bExtKey ? 1 : 0;
    }
    /* not extended key */
    if (!bExtKey) {
      if (b & 0x80) {
       /* key released */
        b ^= 0x80;
        if (nLastKey == (nFlag | b)) {
          nLastKey |= KEY_RELEASED;
        }
      } else {
        /* key pressed */
        nLastKey = nFlag | b;
      }
    }
  }
}

unsigned short GetScanCode(void) {
VECTPROC Int9Prev;
  nFlag = 0;
  bExtKey = 0;
  nLastKey = 0;
  nPauseNum = 0;
  Int9Prev = getvect(0x09);
  setvect(0x09, Int9Hook);
  do {
    while (!(nLastKey & KEY_RELEASED));
    /* drop higher bits since game doesn't use it anyway */
    nLastKey = (nLastKey ^ KEY_RELEASED) & 0xFF;
    /* a key that game can recognize */
  } while (nLastKey > MAX_KEY);
  setvect(0x09, Int9Prev);
  return(nLastKey);
}

#else

/* WIN32 */
unsigned short GetScanCode(void) {
unsigned short k;
  do {
    k = MapVirtualKey(bioskey(0) >> 8, 0);
    /* fix for Right Shift: got Left Shift and Right Shift still pressed */
    if ((k == 0x2A) && (GetAsyncKeyState(VK_RSHIFT) & 0x8000)) {
      k = 0x36;
    }
    /* a key that game can recognize */
  } while (k > MAX_KEY);
  return(k);
}

#endif

char *GetNameByScanCode(unsigned short code) {
  code &= 0xFF;
  code = (code <= MAX_KEY) ? code : 0;
  return(lKeyNames[code]);
}

/* reserved keys that blocked in Dark Forces */
int IsReservedKey(unsigned short key) {
  return(
    /* 01..0F [Escape..Tab] // Escape, Tab, Backspace, -, +, weapon numbers */
    ((key >= 0x01) && (key <= 0x0F)) ||
    /* 3B..44 [F1..F10] // items & things */
    ((key >= 0x3B) && (key <= 0x44)) ||
    /* 1A, 1B {, } // map layer in F1 selection */
     (key == 0x1A) || (key == 0x1B)  ||
    /* 29     [~] // used in cheat */
     (key == 0x29)
  );
}

#ifdef INT9_TEST
int main(void) {
unsigned short k;
  printf("Press a key...\n");
  k = GetScanCode();
  printf("%04X %s\n", k, GetNameByScanCode(k));
  return(0);
}
#endif
